import json
import logging
from dataclasses import asdict, dataclass
from typing import Any, Dict, Iterator, List, Optional, Union

from toolz import dissoc, pipe
from toolz.curried import keyfilter, map

from ..config.config import ChatModel, EmbeddingModel
from ..config.constants import (
    MAX_CHAT_COMPLETION_RETRY_COUNT,
    SYSTEM,
    TOOL,
    USER,
    InvalidForceToolError,
    MaxRetriesExceededError,
    MissingToolCallMessageError,
    Provider,
)
from ..config.models import get_fallback_model
from ..db.db_models import ASSISTANT, FunctionCall
from ..repository.data_models import ContentItem, ContextMessage


def generate_chat_completion_message(
    chat_model: ChatModel,
    context_messages: List[ContextMessage],
    enable_tools: bool = True,
    force_tool: Optional[str] = None,
    retry_number: int = 0,
) -> Iterator[Dict]:
    """
    Generates a chat completion message.

    tool: Force AI to invoke tool
    """

    if force_tool and not enable_tools:
        logging.error("Force tool requested, but tools are disabled. Ignoring force tool request")
        force_tool = None

    from litellm import completion
    from litellm.exceptions import BadRequestError, InternalServerError, RateLimitError

    if context_messages[-1].role == ASSISTANT:
        if force_tool:
            context_messages.append(
                ContextMessage(
                    role=USER,
                    content=f"User is requesting tool call: {force_tool}",
                    chat_model=chat_model.name,
                )
            )
        else:
            raise ValueError("Assistant message already the most recent message")

    context_message_dicts = pipe(
        context_messages,
        map(asdict),
        map(keyfilter(lambda k: k not in ("id", "created_at", "memory_metadata", "chat_model"))),
        map(lambda d: dissoc(d, "tool_calls") if not d.get("tool_calls") else d),
        list,
    )

    if chat_model.ensure_alternating_roles:
        USER_HIDDEN_PREFIX = "[This is a system message, representing internal thought process of the assistant]"
        for idx, message in enumerate(context_message_dicts):
            assert isinstance(message, Dict)

            if idx == 0:
                assert message["role"] == SYSTEM, f"First message must be a system message, but found: " + message["role"]

            if idx != 0 and message["role"] == SYSTEM:
                message["role"] = USER
                message["content"] = f"{USER_HIDDEN_PREFIX} {message['content']}"

    if enable_tools:
        from ..tools.function_caller import get_function_schemas

        if force_tool:
            tools = get_function_schemas()
            if len(tools) == 0:
                raise InvalidForceToolError(f"Requested tool {force_tool}, but not tools available")
            elif not any(t["function"]["name"] == force_tool for t in tools):
                avaliable_tools = ", ".join([t["function"]["name"] for t in tools])
                raise InvalidForceToolError(f"Requested tool {force_tool} not available. Available tools: {avaliable_tools}")
            else:
                tool_choice = {"type": "function", "function": {"name": force_tool}}
        else:
            tools = get_function_schemas()
            tool_choice = "auto"
    else:
        if force_tool:
            raise ValueError(f"Requested tool {force_tool} but model {chat_model.name} does not support tools")
        else:

            if chat_model.provider == Provider.ANTHROPIC and any(m.role == TOOL for m in context_messages):
                # If tool use is in the context window, anthropic requires tools to be enabled and provided
                from ..system_commands import do_not_use
                from ..tools.function_caller import get_function_schemas

                tool_choice = "auto"
                tools = get_function_schemas([do_not_use])  # type: ignore
            else:
                tool_choice = None
                tools = None

    try:
        completion_kwargs = _build_completion_kwargs(
            model=chat_model,
            messages=context_message_dicts,  # type: ignore
            stream=True,
            tool_choice=tool_choice,
            tools=tools,
        )

        tool_call_accumulator = ToolCallAccumulator(chat_model)
        for chunk in completion(**completion_kwargs):
            if chunk.choices[0].delta.content:  # type: ignore
                yield ContentItem(content=chunk.choices[0].delta.content)  # type: ignore
            if chunk.choices[0].delta.tool_calls:  # type: ignore
                yield from tool_call_accumulator.update(chunk.choices[0].delta.tool_calls)  # type: ignore

    except Exception as e:
        if isinstance(e, BadRequestError):
            if "An assistant message with 'tool_calls' must be followed by tool messages" in str(e):
                raise MissingToolCallMessageError
            else:
                raise e
        elif isinstance(e, InternalServerError) or isinstance(e, RateLimitError):
            if retry_number >= MAX_CHAT_COMPLETION_RETRY_COUNT:
                raise MaxRetriesExceededError()
            else:
                fallback_model = get_fallback_model(chat_model)
                if fallback_model:
                    logging.info(
                        f"Rate limit or internal server error for model {chat_model.name}, falling back to model {fallback_model.name}"
                    )
                    yield from generate_chat_completion_message(
                        fallback_model, context_messages, enable_tools, force_tool, retry_number + 1
                    )
                else:
                    logging.error(f"No fallback model available for {chat_model.name}, aborting")
                    raise e
        else:
            raise e


def query_llm(model: ChatModel, prompt: str, system: str) -> str:
    if not prompt:
        raise ValueError("Prompt cannot be empty")
    return _query_llm(model=model, prompt=prompt, system=system)


def query_llm_with_word_limit(model: ChatModel, prompt: str, system: str, word_limit: int) -> str:
    if not prompt:
        raise ValueError("Prompt cannot be empty")
    return query_llm(
        prompt="\n".join(
            [
                prompt,
                f"Your word limit is {word_limit}. DO NOT EXCEED IT.",
            ]
        ),
        model=model,
        system=system,
    )


def get_embedding(model: EmbeddingModel, text: str) -> List[float]:
    """
    Generate an embedding for the given text using the specified model.

    Args:
        text (str): The input text to generate an embedding for.
        model (str): The name of the embedding model to use.

    Returns:
        List[float]: The generated embedding as a list of floats.
    """
    from litellm import embedding

    if not text:
        raise ValueError("Text cannot be empty")
    embedding_kwargs = {
        "model": model.model,
        "input": [text],
        "caching": model.enable_caching,
        "api_key": model.api_key,
    }

    if model.api_base:
        embedding_kwargs["api_base"] = model.api_base
    if model.organization:
        embedding_kwargs["organization"] = model.organization

    response = embedding(**embedding_kwargs)
    return response.data[0]["embedding"]


def _build_completion_kwargs(
    model: ChatModel,
    messages: List[Dict[str, str]],
    stream: bool,
    tool_choice: Union[str, Dict, None],
    tools: Optional[List[Dict[str, Any]]],
) -> Dict[str, Any]:
    """Centralized configuration for LLM requests"""
    kwargs = {
        "messages": messages,
        "model": model.name,
        "api_key": model.api_key,
        "caching": model.enable_caching,
        "tool_choice": tool_choice,
        "tools": tools,
    }

    if model.api_base:
        kwargs["api_base"] = model.api_base
    if model.organization:
        kwargs["organization"] = model.organization
    if stream:
        kwargs["stream"] = True

    return kwargs


def _query_llm(model: ChatModel, prompt: str, system: str) -> str:
    from litellm import completion

    messages = [{"role": SYSTEM, "content": system}, {"role": USER, "content": prompt}]
    completion_kwargs = _build_completion_kwargs(
        model=model,
        messages=messages,
        stream=False,
        tool_choice=None,
        tools=None,
    )
    return completion(**completion_kwargs).choices[0].message.content  # type: ignore


class ToolCallAccumulator:
    from litellm.types.utils import ChatCompletionDeltaToolCall

    def __init__(self, chat_model: ChatModel):
        self.chat_model = chat_model
        self.tool_calls: Dict[int, PartialToolCall] = {}
        self.last_updated_index: Optional[int] = None

    def update(self, delta_tool_calls: Optional[List[ChatCompletionDeltaToolCall]]) -> Iterator[FunctionCall]:
        for delta in delta_tool_calls or []:
            if delta.index not in self.tool_calls:
                if (
                    self.last_updated_index is not None
                    and self.last_updated_index in self.tool_calls
                    and self.last_updated_index != delta.index
                ):
                    raise ValueError("New tool call started, but old one is not yet complete")
                assert delta.id
                self.tool_calls[delta.index] = PartialToolCall(id=delta.id, model=self.chat_model.name)

            completed_tool_call = self.tool_calls[delta.index].update(delta)
            if completed_tool_call:
                self.tool_calls.pop(delta.index)
                yield completed_tool_call
            else:
                self.last_updated_index = delta.index


@dataclass
class PartialToolCall:
    id: str
    model: str  # Add model parameter to determine format
    function_name: str = ""
    arguments: str = ""
    type: str = "function"
    is_complete: bool = False

    from litellm.types.utils import ChatCompletionDeltaToolCall

    def update(self, delta: ChatCompletionDeltaToolCall) -> Optional[FunctionCall]:
        from litellm.types.utils import ChatCompletionDeltaToolCall

        if self.is_complete:
            raise ValueError("PartialToolCall is already complete")

        assert isinstance(delta, ChatCompletionDeltaToolCall), f"Expected ChoiceDeltaToolCall, got {type(delta)}"
        assert delta.function
        if delta.function.name:
            self.function_name += delta.function.name
        if delta.function.arguments:
            self.arguments += delta.function.arguments

        # Check if we have a complete JSON object for arguments
        try:
            function_call = FunctionCall(
                id=self.id,
                function_name=self.function_name,
                arguments=json.loads(self.arguments),
            )
            self.is_complete = True
            return function_call
        except json.JSONDecodeError:
            return None



import logging
from typing import Optional

import typer
from toolz import do

from ..config.constants import UNKNOWN
from ..config.ctx import ElroyContext
from ..db.db_models import UserPreference
from ..utils.utils import is_blank


def set_system_persona(ctx: ElroyContext, system_persona: str) -> str:
    """
    Sets the system instruction for the user
    """
    from ..system_commands import refresh_system_instructions

    system_persona = system_persona.strip()

    if is_blank(system_persona):
        raise typer.BadParameter("System persona cannot be blank.")

    user_preference = get_or_create_user_preference(ctx)

    if user_preference.system_persona == system_persona:
        return do(logging.warning, "New system persona and old system persona are identical")

    user_preference.system_persona = system_persona

    ctx.db.add(user_preference)
    ctx.db.commit()

    refresh_system_instructions(ctx)
    return "System persona updated."


def reset_system_persona(ctx: ElroyContext) -> str:
    """
    Clears the system instruction for the user
    """
    from ..system_commands import refresh_system_instructions

    user_preference = get_or_create_user_preference(ctx)
    if not user_preference.system_persona:
        # Re-clear the persona even if it was already blank, in case some malformed value has been set
        logging.warning("System persona was already set to default")

    user_preference.system_persona = None

    ctx.db.add(user_preference)
    ctx.db.commit()

    refresh_system_instructions(ctx)
    return "System persona cleared, will now use default persona."


def set_user_preferred_name(ctx: ElroyContext, preferred_name: str, override_existing: Optional[bool] = False) -> str:
    """
    Set the user's preferred name. Should predominantly be used relatively early in first conversations, and relatively rarely afterward.

    Args:
        user_id: The user's ID.
        preferred_name: The user's preferred name.
        override_existing: Whether to override the an existing preferred name, if it is already set. Override existing should only be used if a known preferred name has been found to be incorrect.
    """

    user_preference = get_or_create_user_preference(ctx)

    old_preferred_name = user_preference.preferred_name or UNKNOWN

    if old_preferred_name != UNKNOWN and not override_existing:
        return f"Preferred name already set to {user_preference.preferred_name}. If this should be changed, use override_existing=True."
    else:
        user_preference.preferred_name = preferred_name

        ctx.db.commit()
        return f"Set user preferred name to {preferred_name}. Was {old_preferred_name}."


def get_user_preferred_name(ctx: ElroyContext) -> str:
    """Returns the user's preferred name.

    Args:
        user_id (int): the user ID

    Returns:
        str: String representing the user's preferred name.
    """

    user_preference = get_or_create_user_preference(ctx)

    return user_preference.preferred_name or UNKNOWN


def set_user_full_name(ctx: ElroyContext, full_name: str, override_existing: Optional[bool] = False) -> str:
    """Sets the user's full name.

    Guidance for usage:
    - Should predominantly be used relatively in the user journey. However, ensure to not be pushy in getting personal information early.
    - For existing users, this should be used relatively rarely.

    Args:
        user_id (int): user id
        full_name (str): The full name of the user
        override_existing (bool): Whether to override the an existing full name, if it is already set. Override existing should only be used if a known full name has been found to be incorrect.

    Returns:
        str: result of the attempt to set the user's full name
    """

    user_preference = get_or_create_user_preference(ctx)

    old_full_name = user_preference.full_name or UNKNOWN
    if old_full_name != UNKNOWN and not override_existing:
        return f"Full name already set to {user_preference.full_name}. If this should be changed, set override_existing=True."
    else:
        user_preference.full_name = full_name
        ctx.db.commit()

        return f"Full name set to {full_name}. Previous value was {old_full_name}."


def get_user_full_name(ctx: ElroyContext) -> str:
    """Returns the user's full name.

    Args:
        user_id (int): the user ID

    Returns:
        str: String representing the user's full name.
    """

    user_preference = get_or_create_user_preference(ctx)

    return user_preference.full_name or "Unknown name"


def get_or_create_user_preference(ctx: ElroyContext) -> UserPreference:
    from sqlmodel import select

    user_preference = ctx.db.exec(
        select(UserPreference).where(
            UserPreference.user_id == ctx.user_id,
            UserPreference.is_active == True,
        )
    ).first()

    if user_preference is None:
        user_preference = UserPreference(user_id=ctx.user_id, is_active=True)
        ctx.db.add(user_preference)
        ctx.db.commit()
        ctx.db.refresh(user_preference)
    return user_preference


def set_assistant_name(ctx: ElroyContext, assistant_name: str) -> str:
    """
    Sets the assistant name for the user
    """
    from ..system_commands import refresh_system_instructions

    user_preference = get_or_create_user_preference(ctx)
    user_preference.assistant_name = assistant_name
    ctx.db.add(user_preference)
    ctx.db.commit()
    refresh_system_instructions(ctx)
    return f"Assistant name updated to {assistant_name}."



from dataclasses import dataclass, field
from datetime import datetime
from typing import List, NamedTuple, Optional, Union

from ..db.db_models import ASSISTANT, TOOL, FunctionCall, MemoryMetadata, ToolCall


@dataclass
class ContextMessage:
    content: Optional[str]
    role: str
    chat_model: Optional[str]
    id: Optional[int] = None
    created_at: Optional[datetime] = None
    tool_calls: Optional[List[ToolCall]] = None
    tool_call_id: Optional[str] = None
    memory_metadata: List[MemoryMetadata] = field(default_factory=list)

    def __post_init__(self):

        if self.tool_calls is not None:
            self.tool_calls = [ToolCall(**tc) if isinstance(tc, dict) else tc for tc in self.tool_calls]
        # as per openai requirements, empty arrays are disallowed
        if self.tool_calls == []:
            self.tool_calls = None
        if self.role != ASSISTANT and self.tool_calls is not None:
            raise ValueError(f"Only assistant messages can have tool calls, found {self.role} message with tool calls. ID = {self.id}")
        elif self.role != TOOL and self.tool_call_id is not None:
            raise ValueError(f"Only tool messages can have tool call ids, found {self.role} message with tool call id. ID = {self.id}")
        elif self.role == TOOL and self.tool_call_id is None:
            raise ValueError(f"Tool messages must have tool call ids, found tool message without tool call id. ID = {self.id}")


class ContentItem(NamedTuple):
    content: str


StreamItem = Union[ContentItem, FunctionCall]



"""init

Revision ID: 955beece0126
Revises:
Create Date: 2024-12-23 14:46:07.728126

"""

from typing import Sequence, Union

import sqlalchemy as sa
import sqlite_vec
from alembic import op
from sqlmodel.sql.sqltypes import AutoString

# revision identifiers, used by Alembic.
revision: str = "955beece0126"
down_revision: Union[str, None] = None
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:

    # Add debug logging
    import logging

    logging.debug("Starting database upgrade...")

    # Create tables first
    logging.debug("Creating tables...")
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "contextmessageset",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("created_at", sa.DateTime(), nullable=False),
        sa.Column("updated_at", sa.DateTime(), nullable=False),
        sa.Column("user_id", sa.Integer(), nullable=False),
        sa.Column("message_ids", sa.Text(), nullable=True),
        sa.Column("is_active", sa.Boolean(), nullable=True),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("user_id", "is_active"),
    )
    op.create_index(op.f("ix_contextmessageset_id"), "contextmessageset", ["id"], unique=False)
    op.create_table(
        "goal",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("created_at", sa.DateTime(), nullable=False),
        sa.Column("updated_at", sa.DateTime(), nullable=False),
        sa.Column("user_id", sa.Integer(), nullable=False),
        sa.Column("name", AutoString(), nullable=False),
        sa.Column("status_updates", sa.Text(), server_default="[]", nullable=False),
        sa.Column("description", AutoString(), nullable=True),
        sa.Column("strategy", AutoString(), nullable=True),
        sa.Column("end_condition", AutoString(), nullable=True),
        sa.Column("is_active", sa.Boolean(), nullable=True),
        sa.Column("priority", sa.Integer(), nullable=True),
        sa.Column("target_completion_time", sa.DateTime(), nullable=True),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("user_id", "name", "is_active"),
    )
    op.create_index(op.f("ix_goal_id"), "goal", ["id"], unique=False)
    op.create_table(
        "memory",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("created_at", sa.DateTime(), nullable=False),
        sa.Column("updated_at", sa.DateTime(), nullable=False),
        sa.Column("user_id", sa.Integer(), nullable=False),
        sa.Column("name", AutoString(), nullable=False),
        sa.Column("text", AutoString(), nullable=False),
        sa.Column("is_active", sa.Boolean(), nullable=True),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(op.f("ix_memory_id"), "memory", ["id"], unique=False)
    op.create_table(
        "message",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("created_at", sa.DateTime(), nullable=False),
        sa.Column("updated_at", sa.DateTime(), nullable=False),
        sa.Column("user_id", sa.Integer(), nullable=False),
        sa.Column("role", AutoString(), nullable=False),
        sa.Column("content", AutoString(), nullable=True),
        sa.Column("model", AutoString(), nullable=True),
        sa.Column("tool_calls", sa.Text(), nullable=True),
        sa.Column("tool_call_id", AutoString(), nullable=True),
        sa.Column("memory_metadata", sa.Text(), nullable=True),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(op.f("ix_message_id"), "message", ["id"], unique=False)
    op.create_table(
        "user",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("token", AutoString(), nullable=False),
        sa.Column("created_at", sa.DateTime(), nullable=False),
        sa.Column("updated_at", sa.DateTime(), nullable=False),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(op.f("ix_user_id"), "user", ["id"], unique=False)
    op.create_table(
        "userpreference",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("created_at", sa.DateTime(), nullable=False),
        sa.Column("updated_at", sa.DateTime(), nullable=False),
        sa.Column("user_id", sa.Integer(), nullable=False),
        sa.Column("preferred_name", AutoString(), nullable=True),
        sa.Column("system_persona", AutoString(), nullable=True),
        sa.Column("full_name", AutoString(), nullable=True),
        sa.Column("is_active", sa.Boolean(), nullable=True),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("user_id", "is_active"),
    )
    op.create_index(op.f("ix_userpreference_id"), "userpreference", ["id"], unique=False)

    # Then handle vector extension
    logging.debug("Attempting to load sqlite_vec extension...")

    connection = op.get_bind().connection

    try:

        connection.enable_load_extension(True)  # type: ignore
        sqlite_vec.load(connection)  # type: ignore
        connection.enable_load_extension(False)  # type: ignore
    except Exception as e:
        logging.error(f"Failed to load sqlite_vec extension: {e}")
        raise

    logging.debug("Attempting to create vectorstorage table...")

    op.execute(
        """
        CREATE VIRTUAL TABLE IF NOT EXISTS vectorstorage USING vec0(
            id INTEGER PRIMARY KEY,  -- This aliases rowid
            source_type TEXT,
            source_id INTEGER,
            embedding_data FLOAT[1536],
            embedding_text_md5 TEXT
        )
    """
    )

    logging.debug("Database upgrade completed successfully")
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.execute("DROP TABLE IF EXISTS vectorstorage")
    op.drop_index(op.f("ix_userpreference_id"), table_name="userpreference")
    op.drop_table("userpreference")
    op.drop_index(op.f("ix_user_id"), table_name="user")
    op.drop_table("user")
    op.drop_index(op.f("ix_message_id"), table_name="message")
    op.drop_table("message")
    op.drop_index(op.f("ix_memory_id"), table_name="memory")
    op.drop_table("memory")
    op.drop_index(op.f("ix_goal_id"), table_name="goal")
    op.drop_table("goal")
    op.drop_index(op.f("ix_contextmessageset_id"), table_name="contextmessageset")
    op.drop_table("contextmessageset")
    # ### end Alembic commands ###



"""add system instruct to user preference

Revision ID: 71ffabe55f21
Revises: b360a1f1b06e
Create Date: 2024-12-13 21:15:57.911780

"""

from typing import Sequence, Union

import sqlalchemy as sa
from alembic import op
from sqlmodel.sql.sqltypes import AutoString

# revision identifiers, used by Alembic.
revision: str = "71ffabe55f21"
down_revision: Union[str, None] = "b360a1f1b06e"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("userpreference", sa.Column("system_persona", AutoString(), nullable=True))
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("userpreference", "system_persona")
    # ### end Alembic commands ###



import logging
import re
import sqlite3
from pathlib import Path
from struct import unpack
from typing import Iterable, List, Optional, Type

import sqlite_vec
from sqlalchemy import Engine, create_engine, text
from toolz import assoc, pipe
from toolz.curried import map

from ... import PACKAGE_ROOT
from ...config.constants import EMBEDDING_SIZE, RESULT_SET_LIMIT_COUNT
from ..db_manager import DbManager
from ..db_models import EmbeddableSqlModel, VectorStorage


class SqliteManager(DbManager):
    @classmethod
    def _get_config_path(cls):
        return Path(str(PACKAGE_ROOT / "db" / "sqlite" / "alembic" / "alembic.ini"))

    def get_vector_storage_row(self, row: EmbeddableSqlModel) -> Optional[VectorStorage]:
        """Get vector storage entry for a given source type and id"""
        result = self.session.exec(
            text(
                """
                SELECT * FROM vectorstorage
                WHERE source_type = :source_type AND source_id = :source_id
            """
            ).bindparams(
                source_type=row.__class__.__name__, source_id=row.id
            )  # type: ignore
        ).first()

        if result is None:
            return None

        # Convert row to VectorStorage instance
        return pipe(
            dict(result._mapping),  # Convert SQLAlchemy Row to dict
            lambda d: assoc(d, "embedding_data", self._deserialize_embedding(d["embedding_data"])),
            VectorStorage.model_validate,
        )  # type: ignore

    def update_embedding(self, vector_storage: VectorStorage, embedding: List[float], embedding_text_md5: str):
        # Use sqlite_vec's serialize_float32 to properly format the vector data
        serialized_vector = sqlite_vec.serialize_float32(embedding)

        # Use raw SQL with proper parameter binding
        self.session.exec(
            text(
                """
                UPDATE vectorstorage
                SET embedding_data = :embedding_data,
                    embedding_text_md5 = :embedding_text_md5
                WHERE source_type = :source_type
                AND source_id = :source_id
            """
            ).bindparams(
                embedding_data=serialized_vector,
                embedding_text_md5=embedding_text_md5,
                source_type=vector_storage.source_type,
                source_id=vector_storage.source_id,
            )  # type: ignore
        )
        self.session.commit()

    def insert_embedding(self, row: EmbeddableSqlModel, embedding_data, embedding_text_md5):
        # Use sqlite_vec's serialize_float32 to properly format the vector data

        row_id = row.id
        assert row_id

        self.session.exec(
            text(
                """
                INSERT INTO vectorstorage
                (source_type, source_id, embedding_data, embedding_text_md5)
                VALUES
                (:source_type, :source_id, :embedding_data, :embedding_text_md5)
            """
            ).bindparams(
                source_type=row.__class__.__name__,
                source_id=row_id,
                embedding_data=sqlite_vec.serialize_float32(embedding_data),
                embedding_text_md5=embedding_text_md5,
            )  # type: ignore
        )
        self.session.commit()

    def get_embedding(self, row: EmbeddableSqlModel) -> Optional[List[float]]:
        result = self.session.exec(
            text(
                """
                SELECT embedding_data
                FROM vectorstorage
                WHERE source_id = :source_id
                AND source_type = :source_type
            """
            ).bindparams(
                source_id=row.id, source_type=row.__class__.__name__
            )  # type: ignore
        ).first()

        if result is None:
            return None

        # Deserialize the binary data into a list of floats
        return self._deserialize_embedding(result[0])

    def _deserialize_embedding(self, data: bytes) -> List[float]:
        """Deserialize binary vector data from SQLite into a list of floats"""
        return list(unpack(f"{EMBEDDING_SIZE}f", data))

    def query_vector(
        self, l2_distance_threshold: float, table: Type[EmbeddableSqlModel], user_id: int, query: List[float]
    ) -> Iterable[EmbeddableSqlModel]:

        # Serialize the vector once
        serialized_query = sqlite_vec.serialize_float32(query)

        # Raw SQL query
        sql = f"""
        SELECT {table.__tablename__}.*, vec_distance_L2(VectorStorage.embedding_data, ?) as distance
        FROM {table.__tablename__}
        JOIN vectorstorage ON vectorstorage.source_type = ? AND vectorstorage.source_id = {table.__tablename__}.id
        WHERE {table.__tablename__}.user_id = ?
        AND {table.__tablename__}.is_active = 1
        AND vec_distance_L2(VectorStorage.embedding_data, ?) < ?
        ORDER BY distance
        LIMIT ?
        """

        results = self.session.exec(
            text(
                f"""
                SELECT {table.__tablename__}.*, vec_distance_L2(vectorstorage.embedding_data, :query_vec) as distance
                FROM {table.__tablename__}
                JOIN vectorstorage ON vectorstorage.source_type = :source_type
                    AND vectorstorage.source_id = {table.__tablename__}.id
                WHERE {table.__tablename__}.user_id = :user_id
                AND {table.__tablename__}.is_active = 1
                AND vec_distance_L2(vectorstorage.embedding_data, :query_vec) < :threshold
                ORDER BY distance
                LIMIT :limit
            """
            ).bindparams(
                query_vec=serialized_query,
                source_type=table.__name__,
                user_id=user_id,
                threshold=l2_distance_threshold,
                limit=RESULT_SET_LIMIT_COUNT,
            )  # type: ignore
        )

        return pipe(
            results,
            map(lambda row: dict(row._mapping)),  # Convert SQLAlchemy Row to dict
            map(table.model_validate),  # Convert dict to model instance
            list,
            iter,
        )

    @property
    def vector_distance_operator(self) -> str:
        return "l2_distance"

    @classmethod
    def is_valid_url(cls, url):
        pattern = r"^sqlite:\/\/"  # Protocol
        pattern += r"(?:\/)?"  # Optional extra slash for Windows absolute paths
        pattern += r"(?:"  # Start of non-capturing group for alternatives
        pattern += r":memory:|"  # In-memory database option
        pattern += r"\/[^?]+"  # Path to database file
        pattern += r")"  # End of alternatives group
        pattern += r"(?:\?[^#]+)?$"  # Query parameters (optional)
        return bool(re.match(pattern, url))

    @classmethod
    def get_engine(cls, url: str) -> Engine:
        def _sqlite_connect(url):
            # Strip sqlite:/// prefix if present
            db_path = url.replace("sqlite:///", "")
            conn = sqlite3.connect(db_path)
            logging.debug(f"SQLite version: {sqlite3.sqlite_version}")  # Shows SQLite version

            logging.debug("Loading vec extension")
            conn.enable_load_extension(True)
            sqlite_vec.load(conn)
            conn.enable_load_extension(False)
            logging.debug("Vec extension loaded, verifying hello world vector query")
            # Let's verify the function exists after loading
            try:
                conn.execute(
                    "SELECT vec_distance_L2(?, ?)",
                    (sqlite_vec.serialize_float32([0.0]), sqlite_vec.serialize_float32([0.0])),
                )
            except sqlite3.OperationalError as e:
                logging.debug(f"Failed to verify vec_distance_L2 function: {e}")
                raise
            logging.debug("Connection vec extension enabled and verified")
            return conn

        if not cls.is_valid_url(url):
            raise ValueError(f"Invalid database URL: {url}")

        return create_engine(
            url,
            creator=lambda: _sqlite_connect(url),
        )



import json
from functools import partial
from typing import List, Optional, Sequence, Tuple

from sqlmodel import select
from toolz import pipe
from toolz.curried import do, map

from ...config.config import ChatModel
from ...config.ctx import ElroyContext
from ...db.db_models import EmbeddableSqlModel, Memory
from ...llm.client import query_llm
from ..data_models import ContextMessage
from .consolidation import memory_consolidation_check


def get_active_memories(ctx: ElroyContext) -> List[Memory]:
    """Fetch all active memories for the user"""
    return list(
        ctx.db.exec(
            select(Memory).where(
                Memory.user_id == ctx.user_id,
                Memory.is_active == True,
            )
        ).all()
    )


def create_consolidated_memory(ctx: ElroyContext, name: str, text: str, sources: Sequence[EmbeddableSqlModel]):
    from ...repository.embeddings import upsert_embedding_if_needed
    from ..embeddable import add_to_context

    memory = pipe(
        sources,
        map(lambda x: {"source_id": x.id, "source_type": x.__class__.__name__}),
        list,
        json.dumps,
        lambda x: Memory(user_id=ctx.user_id, name=name, text=text, source_metadata=x),
        do(ctx.db.add),
        do(lambda x: ctx.db.commit()),
        do(ctx.db.refresh),
        do(partial(upsert_embedding_if_needed, ctx)),
        do(partial(add_to_context, ctx)),
    )

    [mark_inactive(ctx, m) for m in sources]
    memory_id = memory.id
    assert memory_id
    return memory_id


@memory_consolidation_check
def create_memory(ctx: ElroyContext, name: str, text: str) -> int:
    """Creates a new memory for the assistant.

    Examples of good and bad memory titles are below. Note, the BETTER examples, some titles have been split into two.:

    BAD:
    - [User Name]'s project progress and personal goals: 'Personal goals' is too vague, and the title describes two different topics.

    BETTER:
    - [User Name]'s project on building a treehouse: More specific, and describes a single topic.
    - [User Name]'s goal to be more thoughtful in conversation: Describes a specific goal.

    BAD:
    - [User Name]'s weekend plans: 'Weekend plans' is too vague, and dates must be referenced in ISO 8601 format.

    BETTER:
    - [User Name]'s plan to attend a concert on 2022-02-11: More specific, and includes a specific date.

    BAD:
    - [User Name]'s preferred name and well being: Two different topics, and 'well being' is too vague.

    BETTER:
    - [User Name]'s preferred name: Describes a specific topic.
    - [User Name]'s feeling of rejuvenation after rest: Describes a specific topic.

    Args:
        context (ElroyContext): _description_
        name (str): The name of the memory. Should be specific and discuss one topic.
        text (str): The text of the memory.

    Returns:
        int: The database ID of the memory.
    """
    from ...repository.embeddings import upsert_embedding_if_needed
    from ..embeddable import add_to_context

    memory = Memory(user_id=ctx.user_id, name=name, text=text)
    ctx.db.add(memory)
    ctx.db.commit()
    ctx.db.refresh(memory)

    memory_id = memory.id
    assert memory_id

    upsert_embedding_if_needed(ctx, memory)
    add_to_context(ctx, memory)

    return memory_id


MAX_MEMORY_LENGTH = 12000  # Characters


def manually_record_user_memory(ctx: ElroyContext, text: str, name: Optional[str] = None) -> None:
    """Manually record a memory for the user.

    Args:
        context (ElroyContext): The context of the user.
        name (str): The name of the memory. Should be specific and discuss one topic.
        text (str): The text of the memory.
    """

    if not text:
        raise ValueError("Memory text cannot be empty.")

    if len(text) > MAX_MEMORY_LENGTH:
        raise ValueError(f"Memory text exceeds maximum length of {MAX_MEMORY_LENGTH} characters.")

    if not name:
        name = query_llm(
            ctx.chat_model,
            system="Given text representing a memory, your task is to come up with a short title for a memory. "
            "If the title mentions dates, it should be specific dates rather than relative ones.",
            prompt=text,
        )

    create_memory(ctx, name, text)


async def formulate_memory(
    chat_model: ChatModel, user_preferred_name: Optional[str], context_messages: List[ContextMessage]
) -> Tuple[str, str]:
    from ...llm.prompts import summarize_for_memory
    from ...messaging.context import format_context_messages

    return await summarize_for_memory(
        chat_model,
        format_context_messages(context_messages, user_preferred_name),
        user_preferred_name,
    )


def mark_inactive(ctx: ElroyContext, item: EmbeddableSqlModel):
    from ..embeddable import remove_from_context

    item.is_active = False
    ctx.db.add(item)
    ctx.db.commit()
    remove_from_context(ctx, item)



import logging
from functools import wraps
from typing import Callable, TypeVar

T = TypeVar("T")


def experimental(func: Callable) -> Callable:
    @wraps(func)
    def wrapper(*args, **kwargs):
        ctx = next((arg for arg in args if hasattr(arg, "io")), None)
        if not ctx:
            ctx = next((value for value in kwargs.values() if hasattr(value, "io")), None)

        if ctx and hasattr(ctx, "io"):
            io = ctx.io
            from ..io.base import ElroyIO

            assert isinstance(io, ElroyIO)
            io.notify_warning("Warning: This is an experimental feature.")
        else:
            logging.warning("No context found to notify of experimental feature.")
        return func(*args, **kwargs)

    return wrapper



from pathlib import Path

from platformdirs import user_cache_dir, user_data_dir

APP_NAME = "elroy"
APP_AUTHOR = "elroy-bot"


def get_elroy_home() -> Path:
    """Get the Elroy home directory, creating it if it doesn't exist.

    Returns platform-appropriate path:
    - Windows: C:\\Users\\<username>\\AppData\\Local\\elroy-bot\\elroy
    - macOS: ~/Library/Application Support/elroy
    - Linux: ~/.local/share/elroy
    """
    data_dir = Path(user_data_dir(APP_NAME))
    data_dir.mkdir(parents=True, exist_ok=True)
    return data_dir


def get_default_config_path() -> Path:
    return get_elroy_home() / "elroy.conf.yaml"


def get_elroy_cache() -> Path:
    """Get the Elroy cache directory, creating it if it doesn't exist.

    Returns platform-appropriate path:
    - Windows: C:\\Users\\<username>\\AppData\\Local\\elroy-bot\\elroy\\Cache
    - macOS: ~/Library/Caches/elroy
    - Linux: ~/.cache/elroy
    """
    cache_dir = Path(user_cache_dir(APP_NAME))
    cache_dir.mkdir(parents=True, exist_ok=True)
    return cache_dir


def get_prompt_history_path():
    return get_elroy_cache() / ".history"


def get_default_sqlite_url():
    return f"sqlite:///{get_elroy_cache() / 'elroy.db'}"



import os
from pathlib import Path
from typing import Optional

SQLITE_EXTENSIONS = {".db", ".sqlite", ".sqlite3", ".db3"}


def path_to_sqlite_url(path_str: str) -> Optional[str]:
    """
    Check if a string represents a valid file path with a SQLite extension
    and convert it to a SQLite URL if it is.

    Valid inputs:
    - /absolute/path/to/database.db
    - relative/path/to/database.sqlite
    - C:\\path\\to\\database.db3
    - :memory:

    Args:
        path_str: A string that might represent a file path

    Returns:
        Optional[str]: SQLite URL if path is valid with correct extension, None otherwise
    """
    # Handle memory database special case
    if path_str == ":memory:":
        return "sqlite:///:memory:"

    try:
        # Convert to Path object to validate and normalize
        path = Path(path_str)

        # Check file extension
        if path.suffix.lower() not in SQLITE_EXTENSIONS:
            return None

        # Check if parent directory exists for absolute paths
        if path.is_absolute() and not path.parent.exists():
            return None

        # Convert to forward slashes and ensure proper formatting
        normalized_path = str(path).replace(os.sep, "/")

        # Ensure path starts with / for absolute paths
        if not normalized_path.startswith("/"):
            normalized_path = "/" + normalized_path

        return f"sqlite:///{normalized_path}"

    except (ValueError, RuntimeError):
        return None



"""add memory ops table

Revision ID: 1f8b94fca61b
Revises: 70bdc4f10e11
Create Date: 2025-01-02 19:29:45.907070

"""

from typing import Sequence, Union

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision: str = "1f8b94fca61b"
down_revision: Union[str, None] = "70bdc4f10e11"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "memoryoperationtracker",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("user_id", sa.Integer(), nullable=False),
        sa.Column("memories_since_consolidation", sa.Integer(), nullable=False),
        sa.Column("created_at", sa.DateTime(), nullable=False),
        sa.Column("updated_at", sa.DateTime(), nullable=False),
        sa.PrimaryKeyConstraint("id"),
    )


def downgrade() -> None:
    op.drop_table("memoryoperationtracker")
    # ### end Alembic commands ###



"""add source metadata column

Revision ID: 0fe803cc98f1
Revises: 7781f460cde2
Create Date: 2025-01-03 15:03:21.787143

"""

from typing import Sequence, Union

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision: str = "0fe803cc98f1"
down_revision: Union[str, None] = "7781f460cde2"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("memory", sa.Column("source_metadata", sa.Text(), nullable=True))
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("memory", "source_metadata")
    # ### end Alembic commands ###



"""add assistant name

Revision ID: 66a25cb7d3ef
Revises: 4ed800abf24a
Create Date: 2024-12-26 17:07:33.257863

"""

from typing import Sequence, Union

import sqlalchemy as sa
from alembic import op
from sqlmodel.sql.sqltypes import AutoString

# revision identifiers, used by Alembic.
revision: str = "66a25cb7d3ef"
down_revision: Union[str, None] = "4ed800abf24a"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###

    op.add_column("userpreference", sa.Column("assistant_name", AutoString(), nullable=True))
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("userpreference", "assistant_name")
    # ### end Alembic commands ###



"""rm vector timestamps

Revision ID: 4ed800abf24a
Revises: 892ce7b9a77b
Create Date: 2024-12-24 20:22:49.843503

"""

from typing import Sequence, Union

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision: str = "4ed800abf24a"
down_revision: Union[str, None] = "892ce7b9a77b"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("vectorstorage", "updated_at")
    op.drop_column("vectorstorage", "created_at")
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("vectorstorage", sa.Column("created_at", postgresql.TIMESTAMP(), autoincrement=False, nullable=False))
    op.add_column("vectorstorage", sa.Column("updated_at", postgresql.TIMESTAMP(), autoincrement=False, nullable=False))
    # ### end Alembic commands ###
